@@ -1,6 +1,5 @@
-# vim: tabstop=4 shiftwidth=4 softtabstop=4
-
 # Copyright 2014 IBM Corporation
+# Copyright 2015-2019 Lenovo
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,14 +13,15 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 #
-# This represents the low layer message framing portion of IPMI
+"""This represents the low layer message framing portion of IPMI"""
 
-import pyghmi.exceptions as exc
 import struct
 import threading
 
+import pyghmi.exceptions as exc
 from pyghmi.ipmi.private import constants
 from pyghmi.ipmi.private import session
+from pyghmi.ipmi.private.util import _monotonic_time
 
 
 class Console(object):
@@ -57,6 +57,8 @@ class Console(object):
         self.activated = False
         self.force_session = force
         self.port = port
+        self.ipmi_session = None
+        self.callgotsession = None
         self.ipmi_session = session.Session(bmc=bmc,
                                             userid=userid,
                                             password=password,
@@ -66,13 +68,19 @@ class Console(object):
         # induce one iteration of the loop, now that we would be
         # prepared for it in theory
         session.Session.wait_for_rsp(0)
+        if self.callgotsession is not None:
+            self._got_session(self.callgotsession)
+            self.callgotsession = None
 
     def _got_session(self, response):
-        """Private function to navigate SOL payload activation
-        """
+        """Private function to navigate SOL payload activation"""
+
         if 'error' in response:
             self._print_error(response['error'])
             return
+        if not self.ipmi_session:
+            self.callgotsession = response
+            return
         # Send activate sol payload directive
         # netfn= 6 (application)
         # command = 0x48 (activate payload)
@@ -173,25 +181,25 @@ class Console(object):
             if isinstance(data, dict):
                 self.pendingoutput.append(data)
             else:  # it is a text situation
-                if (len(self.pendingoutput) == 0 or
-                        isinstance(self.pendingoutput[-1], dict)):
+                if (len(self.pendingoutput) == 0
+                        or isinstance(self.pendingoutput[-1], dict)):
                     self.pendingoutput.append(data)
                 else:
                     self.pendingoutput[-1] += data
 
     def _got_cons_input(self, handle):
-        """Callback for handle events detected by ipmi session
-        """
+        """Callback for handle events detected by ipmi session"""
+
         self._addpendingdata(handle.read())
         if not self.awaitingack:
             self._sendpendingoutput()
 
     def close(self):
-        """Shut down an SOL session,
-        """
+        """Shut down an SOL session"""
+
         if self.ipmi_session:
             self.ipmi_session.unregister_keepalive(self.keepaliveid)
-        if self.activated:
+        if self.activated and self.ipmi_session is not None:
             try:
                 self.ipmi_session.raw_command(netfn=6, command=0x49,
                                               data=(1, 1, 0, 0, 0, 0))
@@ -228,23 +236,25 @@ class Console(object):
 
     def _sendpendingoutput(self):
         with self.outputlock:
+            dobreak = False
+            chunk = ''
             if len(self.pendingoutput) == 0:
                 return
             if isinstance(self.pendingoutput[0], dict):
                 if 'break' in self.pendingoutput[0]:
-                    self._sendoutput("", sendbreak=True)
+                    dobreak = True
                 else:
+                    del self.pendingoutput[0]
                     raise ValueError
                 del self.pendingoutput[0]
-                return
-            if len(self.pendingoutput[0]) > self.maxoutcount:
+            elif len(self.pendingoutput[0]) > self.maxoutcount:
                 chunk = self.pendingoutput[0][:self.maxoutcount]
                 self.pendingoutput[0] = self.pendingoutput[0][
                     self.maxoutcount:]
             else:
                 chunk = self.pendingoutput[0]
                 del self.pendingoutput[0]
-            self._sendoutput(chunk)
+            self._sendoutput(chunk, sendbreak=dobreak)
 
     def _sendoutput(self, output, sendbreak=False):
         self.myseq += 1
@@ -257,22 +267,36 @@ class Console(object):
         breakbyte = 0
         if sendbreak:
             breakbyte = 0b10000
-        payload = struct.pack("BBBB", self.myseq, 0, 0, breakbyte)
-        payload += output
+        try:
+            payload = bytearray((self.myseq, 0, 0, breakbyte)) + output
+        except TypeError:  # bytearray hits unicode...
+            payload = bytearray((self.myseq, 0, 0, breakbyte
+                                 )) + output.encode('utf8')
         self.lasttextsize = len(output)
         needskeepalive = False
         if self.lasttextsize == 0:
             needskeepalive = True
         self.awaitingack = True
-        payload = struct.unpack("%dB" % len(payload), payload)
         self.lastpayload = payload
-        self.send_payload(payload, needskeepalive=needskeepalive)
+        self.send_payload(payload, retry=False, needskeepalive=needskeepalive)
+        retries = 5
+        while retries and self.awaitingack:
+            expiry = _monotonic_time() + 5.5 - retries
+            while self.awaitingack and _monotonic_time() < expiry:
+                self.wait_for_rsp(0.5)
+            if self.awaitingack:
+                self.send_payload(payload, retry=False,
+                                  needskeepalive=needskeepalive)
+            retries -= 1
+        if not retries:
+            self._print_error('Connection lost')
 
     def send_payload(self, payload, payload_type=1, retry=True,
                      needskeepalive=False):
         while not (self.connected or self.broken):
             session.Session.wait_for_rsp(timeout=10)
-        if not self.ipmi_session.logged:
+        if self.ipmi_session is None or not self.ipmi_session.logged:
+            self._print_error('Session no longer connected')
             raise exc.IpmiException('Session no longer connected')
         self.ipmi_session.send_payload(payload,
                                        payload_type=payload_type,
@@ -286,8 +310,8 @@ class Console(object):
         self.broken = True
         if self.ipmi_session:
             self.ipmi_session.unregister_keepalive(self.keepaliveid)
-            if (self.ipmi_session.sol_handler and
-                    self.ipmi_session.sol_handler.__self__ is self):
+            if (self.ipmi_session.sol_handler
+                    and self.ipmi_session.sol_handler.__self__ is self):
                 self.ipmi_session.sol_handler = None
             self.ipmi_session = None
         if type(error) == dict:
@@ -305,8 +329,8 @@ class Console(object):
         self.out_handler(data)
 
     def _got_sol_payload(self, payload):
-        """SOL payload callback
-        """
+        """SOL payload callback"""
+
         # TODO(jbjohnso) test cases to throw some likely scenarios at functions
         # for example, retry with new data, retry with no new data
         # retry with unexpected sequence number
@@ -331,10 +355,10 @@ class Console(object):
                 remdatalen = len(payload[4:])  # store remote len before dupe
                 # retry logic, we must ack *this* many even if it is
                 # a retry packet with new partial data
-                remdata = struct.pack("%dB" % remdatalen, *payload[4:])
+                remdata = bytes(payload[4:])
             if newseq == self.remseq:  # it is a retry, but could have new data
                 if remdatalen > self.lastsize:
-                    remdata = remdata[4 + self.lastsize:]
+                    remdata = bytes(remdata[4 + self.lastsize:])
                 else:  # no new data...
                     remdata = ""
             else:  # TODO(jbjohnso) what if remote sequence number is wrong??
@@ -342,7 +366,7 @@ class Console(object):
             self.lastsize = remdatalen
             if remdata:  # Do not subject callers to empty data
                 self._print_data(remdata)
-            ackpayload = (0, self.remseq, remdatalen, 0)
+            ackpayload = bytearray((0, self.remseq, remdatalen, 0))
             # Why not put pending data into the ack? because it's rare
             # and might be hard to decide what to do in the context of
             # retry situation
@@ -363,10 +387,10 @@ class Console(object):
                 else:  # retry all or part of packet, but in a new form
                     # also add pending output for efficiency and ease
                     newtext = self.lastpayload[4 + ackcount:]
-                    newtext = struct.pack("B"*len(newtext), *newtext)
                     with self.outputlock:
-                        if (self.pendingoutput and
-                                not isinstance(self.pendingoutput[0], dict)):
+                        if (self.pendingoutput
+                                and not isinstance(self.pendingoutput[0],
+                                                   dict)):
                             self.pendingoutput[0] = \
                                 newtext + self.pendingoutput[0]
                         else:
@@ -382,7 +406,7 @@ class Console(object):
             # try to mitigate by avoiding overeager retries
             # occasional retry of a packet
             # sooner than timeout suggests is evidently a big deal
-            self.send_payload(payload=self.lastpayload)
+            self.send_payload(payload=self.lastpayload, retry=False)
 
     def main_loop(self):
         """Process all events until no more sessions exist.
@@ -432,8 +456,8 @@ class ServerConsole(Console):
         session.Session.wait_for_rsp(0)
 
     def _got_sol_payload(self, payload):
-        """SOL payload callback
-        """
+        """SOL payload callback"""
+
         # TODO(jbjohnso) test cases to throw some likely scenarios at functions
         # for example, retry with new data, retry with no new data
         # retry with unexpected sequence number
@@ -461,16 +485,16 @@ class ServerConsole(Console):
                 remdatalen = len(payload[4:])  # store remote len before dupe
                 # retry logic, we must ack *this* many even if it is
                 # a retry packet with new partial data
-                remdata = struct.pack("%dB" % remdatalen, *payload[4:])
+                remdata = bytes(payload[4:])
             if newseq == self.remseq:  # it is a retry, but could have new data
                 if remdatalen > self.lastsize:
-                    remdata = remdata[4 + self.lastsize:]
+                    remdata = bytes(remdata[4 + self.lastsize:])
                 else:  # no new data...
                     remdata = ""
             else:  # TODO(jbjohnso) what if remote sequence number is wrong??
                 self.remseq = newseq
             self.lastsize = remdatalen
-            ackpayload = (0, self.remseq, remdatalen, flag)
+            ackpayload = bytearray((0, self.remseq, remdatalen, flag))
             # Why not put pending data into the ack? because it's rare
             # and might be hard to decide what to do in the context of
             # retry situation
@@ -486,10 +510,9 @@ class ServerConsole(Console):
             self.awaitingack = False
             if nacked and not breakdetected:  # the BMC was in some way unhappy
                 newtext = self.lastpayload[4 + ackcount:]
-                newtext = struct.pack("B"*len(newtext), *newtext)
                 with self.outputlock:
-                    if (self.pendingoutput and
-                            not isinstance(self.pendingoutput[0], dict)):
+                    if (self.pendingoutput
+                            and not isinstance(self.pendingoutput[0], dict)):
                         self.pendingoutput[0] = newtext + self.pendingoutput[0]
                     else:
                         self.pendingoutput = [newtext] + self.pendingoutput
@@ -516,6 +539,6 @@ class ServerConsole(Console):
                                        needskeepalive=needskeepalive)
 
     def close(self):
-        """Shut down an SOL session,
-        """
+        """Shut down an SOL session"""
+
         self.activated = False
