@@ -21,10 +21,10 @@ import ast
 import random
 import struct
 
-import pyghmi.exceptions as pygexc
-from pyghmi.ipmi.oem.lenovo import EfiDecompressor
 import six
 
+import pyghmi.exceptions as pygexc
+
 try:
     import EfiCompressor
     from lxml import etree
@@ -36,7 +36,7 @@ IMM_NETFN = 0x2e
 IMM_COMMAND = 0x90
 LENOVO_ENTERPRISE = [0x4d, 0x4f, 0x00]
 
-OPEN_RO_COMMAND = [0x01, 0x05, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10]
+OPEN_RO_COMMAND = [0x01, 0x05, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40]
 OPEN_WO_COMMAND = [0x01, 0x03, 0x01]
 READ_COMMAND = [0x02]
 WRITE_COMMAND = [0x03]
@@ -44,8 +44,14 @@ CLOSE_COMMAND = [0x05]
 SIZE_COMMAND = [0x06]
 
 
+def fromstring(inputdata):
+    if b'!entity' in inputdata.lower():
+        raise Exception('Unsupported XML')
+    return etree.fromstring(inputdata)
+
+
 def run_command_with_retry(connection, data):
-    tries = 10
+    tries = 240
     while tries:
         tries -= 1
         try:
@@ -159,7 +165,7 @@ class LenovoFirmwareConfig(object):
             data.extend(OPEN_WO_COMMAND)
             hex_size = struct.pack("<I", size)
             data.extend(bytearray(hex_size[:4]))
-            data.extend([0x01, 0x10])
+            data.extend([0x01, 0x40])
         if not isinstance(filename, bytes):
             filename = filename.encode('utf-8')
         data.extend(filename)
@@ -186,7 +192,11 @@ class LenovoFirmwareConfig(object):
 
         hex_filehandle = struct.pack("<I", filehandle)
         data.extend(bytearray(hex_filehandle[:4]))
-        run_command_with_retry(self.connection, data=data)
+        try:
+            run_command_with_retry(self.connection, data=data)
+        except pygexc.IpmiException as e:
+            if e.ipmicode != 203:
+                raise
 
     def imm_write(self, filehandle, size, inputdata):
         blocksize = 0xc8
@@ -210,6 +220,7 @@ class LenovoFirmwareConfig(object):
             remaining -= blocksize
             offset += blocksize
             run_command_with_retry(self.connection, data=data)
+            self.connection.ipmi_session.pause(0)
 
     def imm_read(self, filehandle, size):
         blocksize = 0xc8
@@ -232,10 +243,9 @@ class LenovoFirmwareConfig(object):
             data.extend(hex_blocksize[:2])
             remaining -= blocksize
             offset += blocksize
-
             response = run_command_with_retry(self.connection, data=data)
             output += response['data'][5:]
-
+            self.connection.ipmi_session.pause(0)
         return output
 
     def factory_reset(self):
@@ -259,13 +269,13 @@ class LenovoFirmwareConfig(object):
             size = self.imm_size(cfgfilename)
             data = self.imm_read(filehandle, size)
             self.imm_close(filehandle)
-            data = EfiDecompressor.decompress(data)
+            data = EfiCompressor.FrameworkDecompress(data, len(data))
             if len(data) != 0:
                 break
             self.connection.ipmi_session.pause(2)
         if not data:
             raise Exception("BMC failed to return configuration information")
-        xml = etree.fromstring(data)
+        xml = fromstring(data)
         sortid = 0
         for config in xml.iter("config"):
             lenovo_id = config.get("ID")
@@ -462,7 +472,7 @@ class LenovoFirmwareConfig(object):
 
         return options
 
-    def set_fw_options(self, options):
+    def set_fw_options(self, options, checkonly=False):
         changes = False
         random.seed()
         ident = 'ASU-%x-%x-%x-0' % (random.getrandbits(48),
@@ -529,11 +539,18 @@ class LenovoFirmwareConfig(object):
                 choice.append(instance)
 
         if not changes:
-            return
+            return False
+        if checkonly:
+            return True
 
         xml = etree.tostring(configurations)
         data = EfiCompressor.FrameworkCompress(xml, len(xml))
         filehandle = self.imm_open("asu_update.efi", write=True,
                                    size=len(data))
         self.imm_write(filehandle, len(data), data)
+        stubread = len(data)
+        if stubread > 8:
+            stubread = 8
+        self.imm_read(filehandle, stubread)
         self.imm_close(filehandle)
+        return True
