@@ -30,6 +30,7 @@ from cryptography.hazmat.backends import default_backend
 from cryptography.hazmat.primitives.ciphers import algorithms
 from cryptography.hazmat.primitives.ciphers import Cipher
 from cryptography.hazmat.primitives.ciphers import modes
+
 import pyghmi.exceptions as exc
 from pyghmi.ipmi.private import constants
 from pyghmi.ipmi.private import util
@@ -320,7 +321,7 @@ class Session(object):
             for portent in list(cls.bmc_handlers[sesskey]):
                 session = cls.bmc_handlers[sesskey][portent]
                 session.cleaningup = True
-                session.logout()
+                session.logout(False)
 
     @classmethod
     def _assignsocket(cls, server=None, forbiddensockets=()):
@@ -1579,6 +1580,7 @@ class Session(object):
                 # RAKP4 got dropped, BMC can consider it done and we must
                 # restart
                 self._relog()
+                return
             # ignore 15 value if we are retrying.
             # xCAT did but I can't recall why exactly
             if data[1] == 15 and self.logontries:
@@ -1819,6 +1821,8 @@ class Session(object):
                 # if a code somehow makes duplicate SOL handlers,
                 # this would notify all the handlers rather than just the
                 # last one to take ownership
+                if self._customkeepalives[ka][1] is None:
+                    continue
                 self._customkeepalives[ka][1](
                     {'error': 'Session Disconnected'})
         self._customkeepalives = None
