@@ -1,5 +1,5 @@
 # coding: utf8
-# Copyright 2019 Lenovo
+# Copyright 2021 Lenovo
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -31,6 +31,7 @@ import sys
 import time
 
 from dateutil import tz
+
 import pyghmi.constants as const
 import pyghmi.exceptions as exc
 import pyghmi.redfish.oem.lookup as oem
@@ -55,6 +56,7 @@ boot_devices_write = {
     'network': 'Pxe',
     'pxe': 'Pxe',
     'hd': 'Hdd',
+    'usb': 'Usb',
     'cd': 'Cd',
     'cdrom': 'Cd',
     'optical': 'Cd',
@@ -162,11 +164,11 @@ class SensorReading(object):
 class AttrDependencyHandler(object):
     def __init__(self, dependencies, currsettings, pendingsettings):
         self.dependencymap = {}
-        for dep in dependencies.get('Dependencies', [[]]):
-            if dep['Type'] != 'Map':
-                continue
+        for dep in dependencies.get('Dependencies', [{}]):
             if 'Dependency' not in dep:
                 continue
+            if dep['Type'] != 'Map':
+                continue
             if dep['DependencyFor'] in self.dependencymap:
                 self.dependencymap[
                     dep['DependencyFor']].append(dep['Dependency'])
@@ -517,10 +519,25 @@ class Command(object):
     def sysinfo(self):
         return self._do_web_request(self.sysurl)
 
+    @property
+    def bmcinfo(self):
+        return self._do_web_request(self._bmcurl)
+
     def get_power(self):
         currinfo = self._do_web_request(self.sysurl, cache=False)
         return {'powerstate': str(currinfo['PowerState'].lower())}
 
+    def reseat_bay(self, bay):
+        """Request the reseat of a bay
+
+        Request the enclosure manager to reseat the system in a particular
+        bay.
+
+        :param bay: The bay identifier to reseat
+        :return:
+        """
+        self.oem.reseat_bay(bay)
+
     def set_power(self, powerstate, wait=False):
         if powerstate == 'boot':
             oldpowerstate = self.get_power()['powerstate']
@@ -957,7 +974,7 @@ class Command(object):
         """
 
         # For now, this is a stub, no implementation for redfish currently
-        return {}
+        return self.oem.get_bmc_configuration()
 
     def set_bmc_configuration(self, changeset):
         """Get miscellaneous BMC configuration
@@ -969,8 +986,7 @@ class Command(object):
         """
 
         # For now, this is a stub, no implementation for redfish currently
-        raise exc.UnsupportedFunctionality(
-            'Set BMC configuration not supported in redfish yet')
+        return self.oem.set_bmc_configuration(changeset)
 
     def clear_bmc_configuration(self):
         """Reset BMC to factory default
@@ -1094,6 +1110,10 @@ class Command(object):
                     if fnmatch(attr.lower(), change.lower()):
                         found = True
                         changeset[attr] = changeset[change]
+                    if fnmatch(attr.lower(),
+                               change.replace('.', '_').lower()):
+                        found = True
+                        changeset[attr] = changeset[change]
                 if found:
                     del changeset[change]
         for change in changeset:
@@ -1152,11 +1172,13 @@ class Command(object):
                 netmask = _cidr_to_mask(int(cidr))
             patch['IPv4StaticAddresses'] = [ipinfo]
             ipinfo['Address'] = ipv4_address
+            ipv4_configuration = 'static'
             if netmask:
                 ipinfo['SubnetMask'] = netmask
         if ipv4_gateway:
             patch['IPv4StaticAddresses'] = [ipinfo]
             ipinfo['Gateway'] = ipv4_gateway
+            ipv4_configuration = 'static'
         if ipv4_configuration.lower() == 'dhcp':
             dodhcp = True
             patch['DHCPv4'] = {'DHCPEnabled': True}
@@ -1249,248 +1271,82 @@ class Command(object):
         return fwname, currinf
 
     def get_inventory_descriptions(self, withids=False):
-        yield "System"
-        self._hwnamemap = {}
-        for cpu in self._get_cpu_inventory(True, withids):
-            yield cpu
-        for mem in self._get_mem_inventory(True, withids):
-            yield mem
-        for adp in self._get_adp_inventory(True, withids):
-            yield adp
+        return self.oem.get_inventory_descriptions(withids)
 
     def get_inventory_of_component(self, component):
-        if component.lower() == 'system':
-            sysinfo = {
-                'UUID': self.sysinfo.get('UUID', ''),
-                'Serial Number': self.sysinfo.get('SerialNumber', ''),
-                'Manufacturer': self.sysinfo.get('Manufacturer', ''),
-                'Product Name': self.sysinfo.get('Model', ''),
-                'Model': self.sysinfo.get(
-                    'SKU', self.sysinfo.get('PartNumber', '')),
-            }
-            return sysinfo
-        else:
-            for invpair in self.get_inventory():
-                if invpair[0].lower() == component.lower():
-                    return invpair[1]
+        return self.oem.get_inventory_of_component(component)
 
     def get_inventory(self, withids=False):
-        sysinfo = {
-            'UUID': self.sysinfo.get('UUID', ''),
-            'Serial Number': self.sysinfo.get('SerialNumber', ''),
-            'Manufacturer': self.sysinfo.get('Manufacturer', ''),
-            'Product Name': self.sysinfo.get('Model', ''),
-            'Model': self.sysinfo.get(
-                'SKU', self.sysinfo.get('PartNumber', '')),
-        }
-        yield ('System', sysinfo)
-        self._hwnamemap = {}
-        memurl = self.sysinfo.get('Memory', {}).get('@odata.id', None)
-        cpurl = self.sysinfo.get('Processors', {}).get('@odata.id', None)
-        list(self._do_bulk_requests([memurl, cpurl]))
-        adpurls = self._get_adp_urls()
-        cpurls = self._get_cpu_urls()
-        memurls = self._get_mem_urls()
-        diskurls = self._get_disk_urls()
-        allurls = adpurls + cpurls + memurls + diskurls
-        list(self._do_bulk_requests(allurls))
-        for cpu in self._get_cpu_inventory(withids=withids, urls=cpurls):
-            yield cpu
-        for mem in self._get_mem_inventory(withids=withids, urls=memurls):
-            yield mem
-        for adp in self._get_adp_inventory(withids=withids, urls=adpurls):
-            yield adp
-        for disk in self._get_disk_inventory(withids=withids, urls=diskurls):
-            yield disk
+        return self.oem.get_inventory(withids)
 
-    def _get_disk_inventory(self, onlyname=False, withids=False, urls=None):
-        if not urls:
-            urls = self._get_disk_urls()
-        for inf in self._do_bulk_requests(urls):
-            inf, _ = inf
-            ddata = {
-                'Model': inf.get('Model', None),
-                'Serial Number': inf.get('SerialNumber', None),
-                'Description': inf.get('Name'),
-            }
-            loc = inf.get('PhysicalLocation', {}).get('Info', None)
-            if loc:
-                dname = 'Disk {0}'.format(loc)
-            else:
-                dname = inf.get('Id', 'Disk')
-            yield (dname, ddata)
-
-    def _get_adp_inventory(self, onlyname=False, withids=False, urls=None):
-        if not urls:
-            urls = self._get_adp_urls()
-            if not urls:
-                # No PCIe device inventory, but *maybe* ethernet inventory...
-                aidx = 1
-                for nicinfo in self._get_eth_urls():
-                    nicinfo = self._do_web_request(nicinfo)
-                    nicname = nicinfo.get('Name', None)
-                    nicinfo = nicinfo.get('MACAddress', None)
-                    if not nicname:
-                        nicname = 'NIC'
-                    if nicinfo:
-                        yield (nicname,
-                               {'MAC Address {0}'.format(aidx): nicinfo})
-                        aidx += 1
-                return
-        for inf in self._do_bulk_requests(urls):
-            adpinfo, url = inf
-            aname = adpinfo.get('Name', 'Unknown')
-            if aname in self._hwnamemap:
-                aname = adpinfo.get('Id', aname)
-            if aname in self._hwnamemap:
-                self._hwnamemap[aname] = None
-            else:
-                self._hwnamemap[aname] = (url, self._get_adp_inventory)
-            if onlyname:
-                if withids:
-                    yield aname, adpinfo.get('Id', aname)
+    def get_location_information(self):
+        locationinfo = {}
+        for chassis in self.sysinfo.get('Links', {}).get('Chassis', []):
+            chassisurl = chassis['@odata.id']
+            data = self._do_web_request(chassisurl)
+            locdata = data.get('Location', {})
+            postaladdress = locdata.get('PostalAddress', {})
+            placement = locdata.get('Placement', {})
+            contactinfo = locdata.get('Contacts', [])
+            currval = postaladdress.get('Room', '')
+            if currval:
+                locationinfo['room'] = currval
+            currval = postaladdress.get('Location', '')
+            if currval:
+                locationinfo['location'] = currval
+            currval = postaladdress.get('Building', '')
+            if currval:
+                locationinfo['building'] = currval
+            currval = placement.get('Rack', '')
+            if currval:
+                locationinfo['rack'] = currval
+            for contact in contactinfo:
+                contact = contact.get('ContactName', '')
+                if not contact:
+                    continue
+                if 'contactnames' not in locationinfo:
+                    locationinfo['contactnames'] = [contact]
                 else:
-                    yield aname
-                continue
-            functions = adpinfo.get('Links', {}).get('PCIeFunctions', [])
-            nicidx = 1
-            if withids:
-                yieldinf = {'Id': adpinfo.get('Id', aname)}
-            else:
-                yieldinf = {}
-            funurls = [x['@odata.id'] for x in functions]
-            for fun in self._do_bulk_requests(funurls):
-                funinfo, url = fun
-                yieldinf['PCI Device ID'] = funinfo['DeviceId'].replace('0x',
-                                                                        '')
-                yieldinf['PCI Vendor ID'] = funinfo['VendorId'].replace('0x',
-                                                                        '')
-                yieldinf['PCI Subsystem Device ID'] = funinfo[
-                    'SubsystemId'].replace('0x', '')
-                yieldinf['PCI Subsystem Vendor ID'] = funinfo[
-                    'SubsystemVendorId'].replace('0x', '')
-                yieldinf['Type'] = funinfo['DeviceClass']
-                for nicinfo in funinfo.get('Links', {}).get(
-                        'EthernetInterfaces', []):
-                    nicinfo = self._do_web_request(nicinfo['@odata.id'])
-                    macaddr = nicinfo.get('MACAddress', None)
-                    if macaddr:
-                        yieldinf['MAC Address {0}'.format(nicidx)] = macaddr
-                        nicidx += 1
-            yield aname, yieldinf
+                    locationinfo['contactnames'].append(contact)
+        return locationinfo
 
-    def _get_eth_urls(self):
-        ethurls = self.sysinfo.get('EthernetInterfaces', {})
-        ethurls = ethurls.get('@odata.id', None)
-        if ethurls:
-            ethurls = self._do_web_request(ethurls)
-            ethurls = ethurls.get('Members', [])
-            urls = [x['@odata.id'] for x in ethurls]
-        else:
-            urls = []
-        return urls
-
-    def _get_adp_urls(self):
-        adpurls = self.sysinfo.get('PCIeDevices', [])
-        if adpurls:
-            urls = [x['@odata.id'] for x in adpurls]
-        else:
-            urls = []
-        return urls
+    def set_location_information(self, room=None, contactnames=None,
+                                 location=None, building=None, rack=None):
+        locationinfo = {}
+        postaladdress = {}
+        placement = {}
+        if contactnames is not None:
+            locationinfo['Contacts'] = [
+                {'ContactName': x} for x in contactnames]
+        if room is not None:
+            postaladdress['Room'] = room
+        if location is not None:
+            postaladdress['Location'] = location
+        if building is not None:
+            postaladdress['Building'] = building
+        if rack is not None:
+            placement['Rack'] = rack
+        if postaladdress:
+            locationinfo['PostalAddress'] = postaladdress
+        if placement:
+            locationinfo['Placement'] = placement
+        if locationinfo:
+            for chassis in self.sysinfo.get('Links', {}).get('Chassis', []):
+                chassisurl = chassis['@odata.id']
+                self._do_web_request(chassisurl, {'Location': locationinfo},
+                                     method='PATCH')
 
     @property
     def oem(self):
         if not self._oem:
             self._oem = oem.get_oem_handler(
-                self.sysinfo, self.sysurl, self.wc, self._urlcache)
+                self.sysinfo, self.sysurl, self.wc, self._urlcache, self)
             self._oem.set_credentials(self.username, self.password)
         return self._oem
 
     def get_description(self):
         return self.oem.get_description()
 
-    def _get_cpu_inventory(self, onlynames=False, withids=False, urls=None):
-        if not urls:
-            urls = self._get_cpu_urls()
-        if not urls:
-            return
-        for res in self._do_bulk_requests(urls):
-            currcpuinfo, url = res
-            name = currcpuinfo.get('Name', 'CPU')
-            if name in self._hwnamemap:
-                self._hwnamemap[name] = None
-            else:
-                self._hwnamemap[name] = (url, self._get_cpu_inventory)
-            if onlynames:
-                yield name
-                continue
-            cpuinfo = {'Model': currcpuinfo.get('Model', None)}
-            yield name, cpuinfo
-
-    def _get_disk_urls(self):
-        storurl = self.sysinfo.get('Storage', {}).get('@odata.id', None)
-        urls = []
-        if storurl:
-            storurl = self._do_web_request(storurl)
-            for url in storurl.get('Members', []):
-                url = url['@odata.id']
-                ctldata = self._do_web_request(url)
-                for durl in ctldata.get('Drives', []):
-                    urls.append(durl['@odata.id'])
-        return urls
-
-    def _get_cpu_urls(self):
-        cpurl = self.sysinfo.get('Processors', {}).get('@odata.id', None)
-        if cpurl is None:
-            urls = []
-        else:
-            cpurl = self._do_web_request(cpurl)
-            urls = [x['@odata.id'] for x in cpurl.get('Members', [])]
-        return urls
-
-    def _get_mem_inventory(self, onlyname=False, withids=False, urls=None):
-        if not urls:
-            urls = self._get_mem_urls()
-        if not urls:
-            return
-        for mem in self._do_bulk_requests(urls):
-            currmeminfo, url = mem
-            name = currmeminfo.get('Name', 'Memory')
-            if name in self._hwnamemap:
-                self._hwnamemap[name] = None
-            else:
-                self._hwnamemap[name] = (url, self._get_mem_inventory)
-            if onlyname:
-                yield name
-                continue
-            if currmeminfo.get(
-                    'Status', {}).get('State', 'Absent') == 'Absent':
-                yield (name, None)
-                continue
-            currspeed = currmeminfo.get('OperatingSpeedMhz', None)
-            if currspeed:
-                currspeed = int(currspeed)
-                currspeed = currspeed * 8 - (currspeed * 8 % 100)
-            meminfo = {
-                'capacity_mb': currmeminfo.get('CapacityMiB', None),
-                'manufacturer': currmeminfo.get('Manufacturer', None),
-                'memory_type': currmeminfo.get('MemoryDeviceType', None),
-                'model': currmeminfo.get('PartNumber', None),
-                'module_type': currmeminfo.get('BaseModuleType', None),
-                'serial': currmeminfo.get('SerialNumber', None),
-                'speed': currspeed,
-            }
-            yield (name, meminfo)
-
-    def _get_mem_urls(self):
-        memurl = self.sysinfo.get('Memory', {}).get('@odata.id', None)
-        if not memurl:
-            urls = []
-        else:
-            memurl = self._do_web_request(memurl)
-            urls = [x['@odata.id'] for x in memurl.get('Members', [])]
-        return urls
-
     def get_event_log(self, clear=False):
         bmcinfo = self._do_web_request(self._bmcurl)
         lsurl = bmcinfo.get('LogServices', {}).get('@odata.id', None)
@@ -1657,10 +1513,16 @@ class Command(object):
             return
         for vmurl in vmurls:
             vminfo = self._do_web_request(vmurl, cache=False)
-            if vminfo['ConnectedVia'] != 'NotConnected':
+            if vminfo.get('ConnectedVia', None) != 'NotConnected':
                 continue
-            self._do_web_request(vmurl, {'Image': url, 'Inserted': True},
-                                 'PATCH')
+            try:
+                self._do_web_request(vmurl, {'Image': url, 'Inserted': True},
+                                     'PATCH')
+            except exc.RedfishError as re:
+                if re.msgid.endswith(u'PropertyUnknown'):
+                    self._do_web_request(vmurl, {'Image': url}, 'PATCH')
+                else:
+                    raise
             break
 
     def detach_remote_media(self):
@@ -1676,12 +1538,19 @@ class Command(object):
             for vminfo in self._do_bulk_requests(vmurls):
                 vminfo, currl = vminfo
                 if vminfo['Image']:
-                    self._do_web_request(currl,
-                                         {'Image': None,
-                                          'Inserted': False},
-                                         method='PATCH')
+                    try:
+                        self._do_web_request(currl,
+                                             {'Image': None,
+                                              'Inserted': False},
+                                             method='PATCH')
+                    except exc.RedfishError as re:
+                        if re.msgid.endswith(u'PropertyUnknown'):
+                            self._do_web_request(currl, {'Image': None},
+                                                 method='PATCH')
+                        else:
+                            raise
 
-    def upload_media(self, filename, progress=None):
+    def upload_media(self, filename, progress=None, data=None):
         """Upload a file to be hosted on the target BMC
 
         This will upload the specified data to
@@ -1692,7 +1561,7 @@ class Command(object):
                          will be given to the bmc.
         :param progress: Optional callback for progress updates
         """
-        return self.oem.upload_media(filename, progress)
+        return self.oem.upload_media(filename, progress, data)
 
     def update_firmware(self, file, data=None, progress=None, bank=None):
         """Send file to BMC to perform firmware update
@@ -1704,6 +1573,8 @@ class Command(object):
                            update process.  Provide if
          :param bank: Indicate a target 'bank' of firmware if supported
         """
+        if progress is None:
+            progress = lambda x: True
         return self.oem.update_firmware(file, data, progress, bank)
 
     def get_diagnostic_data(self, savefile, progress=None, autosuffix=False):
@@ -1726,8 +1597,8 @@ class Command(object):
                     directory))
         return self.oem.save_licenses(directory)
 
-    def apply_license(self, filename, progress=None):
-        return self.oem.apply_license(filename, progress)
+    def apply_license(self, filename, progress=None, data=None):
+        return self.oem.apply_license(filename, progress, data)
 
 
 if __name__ == '__main__':
