@@ -25,9 +25,10 @@ import socket
 import ssl
 import threading
 
-import pyghmi.exceptions as pygexc
 import six
 
+import pyghmi.exceptions as pygexc
+
 try:
     import Cookie
     import httplib
@@ -140,7 +141,7 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
             self.cert_reqs = ssl.CERT_NONE  # use custom validation
         else:
             self.cert_reqs = ssl.CERT_REQUIRED  # use standard validation
-        if '[' not in host and '%' in host:
+        if '[' not in host and '%' in host and 'Host'not in self.stdheaders:
             self.stdheaders['Host'] = '[' + host[:host.find('%')] + ']'
 
     def dupe(self):
@@ -238,7 +239,12 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
             raise
         body = rsp.read()
         if rsp.getheader('Content-Encoding', None) == 'gzip':
-            body = gzip.GzipFile(fileobj=io.BytesIO(body)).read()
+            try:
+                body = gzip.GzipFile(fileobj=io.BytesIO(body)).read()
+            except IOError:
+                # some implementations will send non-gzipped and claim it as
+                # gzip
+                pass
         if rsp.status >= 200 and rsp.status < 300:
             if body and not isinstance(body, str):
                 body = body.decode('utf8')
@@ -306,7 +312,12 @@ class SecureHTTPConnection(httplib.HTTPConnection, object):
                             % rsp.read())
         body = rsp.read()
         if rsp.getheader('Content-Encoding', None) == 'gzip':
-            body = gzip.GzipFile(fileobj=io.BytesIO(body)).read()
+            try:
+                body = gzip.GzipFile(fileobj=io.BytesIO(body)).read()
+            except IOError:
+                # In case the implementation lied, let the body return
+                # unprocessed
+                pass
         return body
 
     def get_upload_progress(self):
